const App = @import("./App.zig");

pub const colors = @import("./gfx/colors.zig");

const coral = @import("coral");

const bmp = @import("./gfx/bmp.zig");

const ext = @import("./ext.zig");

const flow = @import("flow");

const handles = @import("./gfx/handles.zig");

pub const lina = @import("./gfx/lina.zig");

const msg = @import("./msg.zig");

const spirv = @import("./gfx/spirv.zig");

const rendering = @import("./gfx/rendering.zig");

const std = @import("std");

pub const Assets = struct {
	window: *ext.SDL_Window,
	texture_formats: coral.stack.Sequential(TextureFormat),
	staging_arena: std.heap.ArenaAllocator,
	frame_rendered: std.Thread.ResetEvent = .{},

	pub const LoadError = std.mem.Allocator.Error;

	pub const LoadFileError = LoadError || coral.files.AccessError || error {
		FormatUnsupported,
	};

	pub const TextureFormat = struct {
		extension: []const u8,
		load_file: *const fn (*std.heap.ArenaAllocator, coral.files.Storage, []const u8) LoadFileError!Texture.Desc,
	};

	fn deinit(self: *Assets) void {
		rendering.enqueue_work(.shutdown);
		self.staging_arena.deinit();
		self.texture_formats.deinit();
	}

	fn init() !Assets {
		const window = create: {
			const position = ext.SDL_WINDOWPOS_CENTERED;
			const flags = ext.SDL_WINDOW_OPENGL;
			const width = 640;
			const height = 480;

			break: create ext.SDL_CreateWindow("Ona", position, position, width, height, flags) orelse {
				return error.Unsupported;
			};
		};

		errdefer {
			ext.SDL_DestroyWindow(window);
		}

		try rendering.startup(window);

		return .{
			.staging_arena = std.heap.ArenaAllocator.init(coral.heap.allocator),
			.texture_formats = .{.allocator = coral.heap.allocator},
			.window = window,
		};
	}

	pub fn load_texture(_: *Assets, desc: Texture.Desc) std.mem.Allocator.Error!Texture {
		var loaded = coral.asyncio.Future(std.mem.Allocator.Error!Texture){};

		rendering.enqueue_work(.{
			.load_texture = .{
				.desc = desc,
				.loaded = &loaded,
			},
		});

		return loaded.get().*;
	}

	pub fn load_texture_file(self: *Assets, storage: coral.files.Storage, path: []const u8) LoadFileError!Texture {
		defer {
			const max_cache_size = 536870912;

			if (!self.staging_arena.reset(.{.retain_with_limit = max_cache_size})) {
				std.log.warn("failed to retain staging arena size of {} bytes", .{max_cache_size});
			}
		}

		for (self.texture_formats.values) |format| {
			if (!std.mem.endsWith(u8, path, format.extension)) {
				continue;
			}

			return self.load_texture(try format.load_file(&self.staging_arena, storage, path));
		}

		return error.FormatUnsupported;
	}

	pub const thread_restriction = .main;
};

pub const Commands = struct {
	list: *rendering.Commands.List,

	pub const State = struct {
		commands: *rendering.Commands,
	};

	pub fn bind(_: flow.system.BindContext) std.mem.Allocator.Error!State {
		var created_commands = coral.asyncio.Future(std.mem.Allocator.Error!*rendering.Commands){};

		rendering.enqueue_work(.{
			.create_commands = .{
				.created = &created_commands,
			},
		});

		return .{
			.commands = try created_commands.get().*,
		};
	}

	pub fn init(state: *State) Commands {
		return .{
			.list = state.commands.pending_list(),
		};
	}

	pub fn draw_texture(self: Commands, command: rendering.Command.DrawTexture) std.mem.Allocator.Error!void {
		try self.list.append(.{.draw_texture = command});
	}

	pub fn set_target(self: Commands, command: rendering.Command.SetTarget) std.mem.Allocator.Error!void {
		try self.list.append(.{.set_target = command});
	}
};

pub const Config = struct {
	width: u16 = 1280,
	height: u16 = 720,
	clear_color: lina.Color = colors.black,
};

pub const Input = union (enum) {
	key_up: Key,
	key_down: Key,

	pub const Key = enum (u32) {
		a = ext.SDL_SCANCODE_A,
		d = ext.SDL_SCANCODE_D,
		s = ext.SDL_SCANCODE_S,
		w = ext.SDL_SCANCODE_W,

		pub fn scancode(key: Key) u32 {
			return @intFromEnum(key);
		}
	};
};

pub const Rect = struct {
	left: f32,
	top: f32,
	right: f32,
	bottom: f32,
};

pub const Texture = handles.Texture;

pub fn poll(app: flow.Write(App), inputs: msg.Send(Input)) !void {
	var event = @as(ext.SDL_Event, undefined);

	while (ext.SDL_PollEvent(&event) != 0) {
		switch (event.type) {
			ext.SDL_QUIT => app.res.quit(),
			ext.SDL_KEYUP => try inputs.push(.{.key_up = @enumFromInt(event.key.keysym.scancode)}),
			ext.SDL_KEYDOWN => try inputs.push(.{.key_down = @enumFromInt(event.key.keysym.scancode)}),
			else => {},
		}
	}
}

pub fn setup(world: *flow.World, events: App.Events) (error {Unsupported} || std.Thread.SpawnError || std.mem.Allocator.Error)!void {
	if (ext.SDL_Init(ext.SDL_INIT_VIDEO) != 0) {
		return error.Unsupported;
	}

	const assets = create: {
		var assets = try Assets.init();

		errdefer {
			assets.deinit();
		}

		break: create try world.set_get_resource(assets);
	};

	assets.frame_rendered.set();

	errdefer {
		assets.deinit();
	}

	const builtin_texture_formats = [_]Assets.TextureFormat{
		.{
			.extension = "bmp",
			.load_file = bmp.load_file,
		},
	};

	for (builtin_texture_formats) |format| {
		try assets.texture_formats.push_grow(format);
	}

	try world.set_resource(Config{});
	try world.on_event(events.pre_update, flow.system_fn(poll), .{.label = "poll gfx"});
	try world.on_event(events.exit, flow.system_fn(stop), .{.label = "stop gfx"});
	try world.on_event(events.finish, flow.system_fn(synchronize), .{.label = "synchronize gfx"});
}

pub fn stop(assets: flow.Write(Assets)) void {
	assets.res.deinit();
}

pub fn synchronize(assets: flow.Write(Assets), config: flow.Read(Config)) !void {
	assets.res.frame_rendered.wait();
	assets.res.frame_rendered.reset();

	var commands_swapped = std.Thread.ResetEvent{};

	rendering.enqueue_work(.{
		.rotate_commands = .{
			.finished = &commands_swapped,
		},
	});

	var display_width, var display_height = [_]c_int{0, 0};

	ext.SDL_GL_GetDrawableSize(assets.res.window, &display_width, &display_height);

	if (config.res.width != display_width or config.res.height != display_height) {
		ext.SDL_SetWindowSize(assets.res.window, config.res.width, config.res.height);
	}

	commands_swapped.wait();

	rendering.enqueue_work(.{
		.render_frame = .{
			.width = config.res.width,
			.height = config.res.height,
			.clear_color = config.res.clear_color,
			.finished = &assets.res.frame_rendered,
		},
	});
}
