const coral = @import("coral");

const ext = @import("../ext.zig");

const handles = @import("./handles.zig");

const lina = @import("./lina.zig");

const resources = @import("./resources.zig");

const spirv = @import("./spirv.zig");

const sokol = @import("sokol");

const std = @import("std");

pub const Command = union (enum) {
	draw_texture: DrawTexture,
	set_target: SetTarget,

	pub const DrawTexture = struct {
		texture: handles.Texture,
		transform: lina.Transform2D,
	};

	pub const SetTarget = struct {
		texture: ?handles.Texture = null,
		clear_color: ?lina.Color,
		clear_depth: ?f32,
		clear_stencil: ?u8,
	};

	fn clone(self: Command, arena: *std.heap.ArenaAllocator) !Command {
		_ = arena;

		return switch (self) {
			.draw_texture => |draw_texture| .{.draw_texture = draw_texture},
			.set_target => |set_target| .{.set_target = set_target},
		};
	}

	fn process(self: Command, pools: *Pools, frame: *Frame) !void {
		try switch (self) {
			.draw_texture => |draw_texture| frame.draw_texture(pools, draw_texture),
			.set_target => |set_target| frame.set_target(pools, set_target),
		};
	}
};

pub const Commands = struct {
	swap_lists: [2]List,
	swap_state: u1 = 0,
	has_next: ?*Commands = null,
	has_prev: ?*Commands = null,

	pub const List = struct {
		arena: std.heap.ArenaAllocator,
		stack: coral.stack.Sequential(Command),

		pub fn append(self: *List, command: Command) std.mem.Allocator.Error!void {
			return self.stack.push_grow(try command.clone(&self.arena));
		}

		pub fn clear(self: *List) void {
			self.stack.clear();

			if (!self.arena.reset(.retain_capacity)) {
				std.log.warn("failed to reset the buffer of a gfx queue with retained capacity", .{});
			}
		}

		fn deinit(self: *List) void {
			self.arena.deinit();
			self.stack.deinit();

			self.* = undefined;
		}

		fn init(allocator: std.mem.Allocator) List {
			return .{
				.arena = std.heap.ArenaAllocator.init(allocator),
				.stack = .{.allocator = allocator},
			};
		}
	};

	const Self = @This();

	pub fn deinit(self: *Self) void {
		for (&self.swap_lists) |*list| {
			list.deinit();
		}

		self.* = undefined;
	}

	pub fn init(allocator: std.mem.Allocator) Self {
		return .{
			.swap_lists = .{List.init(allocator), List.init(allocator)},
		};
	}

	pub fn pending_list(self: *Self) *List {
		return &self.swap_lists[self.swap_state];
	}

	fn rotate(self: *Self) void {
		const swapped_state = self.swap_state ^ 1;

		self.swap_lists[swapped_state].clear();

		self.swap_state = swapped_state;
	}

	pub fn submitted_commands(self: *const Self) []const Command {
		return self.swap_lists[self.swap_state ^ 1].stack.values;
	}
};

const Frame = struct {
	swapchain: sokol.gfx.Swapchain,
	drawn_count: usize = 0,
	flushed_count: usize = 0,
	current_source_texture: handles.Texture = .default,
	current_target_texture: ?handles.Texture = null,

	const DrawTexture = extern struct {
		transform: lina.Transform2D,
		tint: @Vector(4, u8) = @splat(std.math.maxInt(u8)),
		depth: f32 = 0,
		texture_offset: @Vector(2, f32) = @splat(0),
		texture_size: @Vector(2, f32) = @splat(1),
	};

	pub fn draw_texture(self: *Frame, pools: *Pools, command: Command.DrawTexture) !void {
		if (command.texture != self.current_source_texture) {
			self.flush(pools);
		}

		self.current_source_texture = command.texture;

		const has_filled_current_buffer = (self.drawn_count % batches_per_buffer) == 0;
		const buffer_count = self.drawn_count / batches_per_buffer;

		if (has_filled_current_buffer and buffer_count == texture_batch_buffers.len()) {
			const instance_buffer = sokol.gfx.makeBuffer(.{
				.size = @sizeOf(DrawTexture) * batches_per_buffer,
				.usage = .STREAM,
			});

			errdefer sokol.gfx.destroyBuffer(instance_buffer);

			try texture_batch_buffers.push_grow(instance_buffer);
		}

		_ = sokol.gfx.appendBuffer(texture_batch_buffers.get().?, sokol.gfx.asRange(&DrawTexture{
			.transform = command.transform,
		}));

		self.drawn_count += 1;
	}

	pub fn flush(self: *Frame, pools: *Pools) void {
		if (self.flushed_count == self.drawn_count) {
			return;
		}

		var bindings = sokol.gfx.Bindings{
			.index_buffer = quad_index_buffer,
		};

		bindings.vertex_buffers[vertex_indices.mesh] = quad_vertex_buffer;

		switch (pools.textures.get(@intFromEnum(self.current_source_texture)).?.*) {
			.render => |render| {
				bindings.fs.images[0] = render.color_image;
				bindings.fs.samplers[0] = render.sampler;
				bindings.vs.images[0] = render.color_image;
				bindings.vs.samplers[0] = render.sampler;
			},

			.static => |static| {
				bindings.fs.images[0] = static.image;
				bindings.fs.samplers[0] = static.sampler;
				bindings.vs.images[0] = static.image;
				bindings.vs.samplers[0] = static.sampler;
			},
		}

		if (self.current_target_texture) |target_texture| {
			const target_view_buffer = pools.textures.get(@intFromEnum(target_texture)).?.render.view_buffer;

			bindings.fs.storage_buffers[0] = target_view_buffer;
			bindings.vs.storage_buffers[0] = target_view_buffer;
		} else {
			bindings.fs.storage_buffers[0] = view_buffer;
			bindings.vs.storage_buffers[0] = view_buffer;
		}

		sokol.gfx.applyPipeline(texture_batching_pipeline);

		while (true) {
			const buffer_index = self.flushed_count / batches_per_buffer;
			const buffer_offset = self.flushed_count % batches_per_buffer;
			const instances_to_flush = @min(batches_per_buffer - buffer_offset, self.drawn_count - self.flushed_count);

			self.flushed_count += instances_to_flush;
			bindings.vertex_buffers[vertex_indices.instance] = texture_batch_buffers.values[buffer_index];
			bindings.vertex_buffer_offsets[vertex_indices.instance] = @intCast(@sizeOf(DrawTexture) * buffer_offset);

			sokol.gfx.applyBindings(bindings);
			sokol.gfx.draw(0, 6, @intCast(instances_to_flush));

			if (self.flushed_count == self.drawn_count) {
				break;
			}
		}
	}

	pub fn set_target(self: *Frame, pools: *Pools, command: Command.SetTarget) void {
		sokol.gfx.endPass();

		var pass = sokol.gfx.Pass{
			.action = .{.stencil = .{.load_action = .CLEAR}},
		};

		if (command.clear_color) |color| {
			pass.action.colors[0] = .{
				.load_action = .CLEAR,
				.clear_value = @bitCast(color),
			};
		} else {
			pass.action.colors[0] = .{.load_action = .LOAD};
		}

		if (command.clear_depth) |depth| {
			pass.action.depth = .{
				.load_action = .CLEAR,
				.clear_value = depth,
			};
		} else {
			pass.action.depth = .{.load_action = .LOAD};
		}

		if (command.texture) |texture| {
			pass.attachments = switch (pools.textures.get(@intFromEnum(texture)).?.*) {
				.static => @panic("Cannot render to static textures"),
				.render => |render| render.attachments,
			};

			self.current_target_texture = command.texture;
		} else {
			pass.swapchain = self.swapchain;
			self.current_target_texture = null;
		}

		sokol.gfx.beginPass(pass);
	}

	const storage_bindings = .{
		.engine = 0,
		.material = 1,
	};

	var texture_batch_buffers = coral.stack.Sequential(sokol.gfx.Buffer){.allocator = coral.heap.allocator};

	const batches_per_buffer = 512;

	var texture_batching_pipeline: sokol.gfx.Pipeline = undefined;
};

const Pools = struct {
	textures: TexturePool,

	const TexturePool = coral.Pool(resources.Texture);

	fn create_texture(self: *Pools, desc: handles.Texture.Desc) !handles.Texture {
		var texture = try resources.Texture.init(desc);

		errdefer texture.deinit();

		return @enumFromInt(try self.textures.insert(texture));
	}

	fn deinit(self: *Pools) void {
		var textures = self.textures.values();

		while (textures.next()) |texture| {
			texture.deinit();
		}

		self.textures.deinit();
	}

	fn destroy_texture(self: *Pools, texture_key: handles.Texture) bool {
		switch (texture_key) {
			.default => {},

			else => {
				var texture = self.textures.remove(@intFromEnum(texture_key)) orelse {
					return false;
				};

				texture.deinit();
			},
		}

		return true;
	}

	fn init(allocator: std.mem.Allocator) !Pools {
		var pools = Pools{
			.textures = TexturePool.init(allocator),
		};

		errdefer pools.deinit();

		const default_texture_data = [_]u32{
			0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF,
			0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF,
			0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF,
			0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF,
			0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF,
			0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF,
			0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF,
			0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF, 0x800080FF, 0x000000FF,
		};

		_ = try pools.create_texture(.{
			.format = .rgba8,

			.access = .{
				.static = .{
					.data = std.mem.asBytes(&default_texture_data),
					.width = 8,
				},
			},
		});

		return pools;
	}
};

pub const Work = union (enum) {
	create_commands: CreateCommandsWork,
	load_texture: LoadTextureWork,
	render_frame: RenderFrameWork,
	rotate_commands: RotateCommandsWork,
	shutdown,
	unload_texture: UnloadTextureWork,

	pub const CreateCommandsWork = struct {
		created: *coral.asyncio.Future(std.mem.Allocator.Error!*Commands),
	};

	pub const LoadTextureWork = struct {
		desc: handles.Texture.Desc,
		loaded: *coral.asyncio.Future(std.mem.Allocator.Error!handles.Texture),
	};

	pub const RotateCommandsWork = struct {
		finished: *std.Thread.ResetEvent,
	};

	pub const RenderFrameWork = struct {
		clear_color: lina.Color,
		width: u16,
		height: u16,
		finished: *std.Thread.ResetEvent,
	};

	pub const UnloadTextureWork = struct {
		handle: handles.Texture,
	};

	var pending: coral.asyncio.BlockingQueue(1024, Work) = .{};
};

pub fn enqueue_work(work: Work) void {
	Work.pending.enqueue(work);
}

pub fn startup(window: *ext.SDL_Window) !void {
	work_thread = try std.Thread.spawn(.{}, run, .{window});
}

var quad_index_buffer: sokol.gfx.Buffer = undefined;

var quad_vertex_buffer: sokol.gfx.Buffer = undefined;

var view_buffer: sokol.gfx.Buffer = undefined;

const vertex_indices = .{
	.mesh = 0,
	.instance = 1,
};

const vertex_layout_state = sokol.gfx.VertexLayoutState{
	.attrs = get: {
		var attrs = [_]sokol.gfx.VertexAttrState{.{}} ** 16;

		attrs[0] = .{
			.format = .FLOAT2,
			.buffer_index = vertex_indices.mesh,
		};

		attrs[1] = .{
			.format = .FLOAT2,
			.buffer_index = vertex_indices.mesh,
		};

		attrs[2] = .{
			.format = .FLOAT2,
			.buffer_index = vertex_indices.instance,
		};

		attrs[3] = .{
			.format = .FLOAT2,
			.buffer_index = vertex_indices.instance,
		};

		attrs[4] = .{
			.format = .FLOAT2,
			.buffer_index = vertex_indices.instance,
		};

		attrs[5] = .{
			.format = .UBYTE4N,
			.buffer_index = vertex_indices.instance,
		};

		attrs[6] = .{
			.format = .FLOAT,
			.buffer_index = vertex_indices.instance,
		};

		attrs[7] = .{
			.format = .FLOAT4,
			.buffer_index = vertex_indices.instance,
		};

		break: get attrs;
	},

	.buffers = get: {
		var vertex_buffer_layouts = [_]sokol.gfx.VertexBufferLayoutState{.{}} ** 8;

		vertex_buffer_layouts[vertex_indices.instance].step_func = .PER_INSTANCE;

		break: get vertex_buffer_layouts;
	},
};

fn run(window: *ext.SDL_Window) !void {
	const context = configure_and_create: {
		var result = @as(c_int, 0);

		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_FLAGS, ext.SDL_GL_CONTEXT_FORWARD_COMPATIBLE_FLAG);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_PROFILE_MASK, ext.SDL_GL_CONTEXT_PROFILE_CORE);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MAJOR_VERSION, 3);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_CONTEXT_MINOR_VERSION, 3);
		result |= ext.SDL_GL_SetAttribute(ext.SDL_GL_DOUBLEBUFFER, 1);

		if (result != 0) {
			return error.Unsupported;
		}

		break: configure_and_create ext.SDL_GL_CreateContext(window);
	};

	defer ext.SDL_GL_DeleteContext(context);

	sokol.gfx.setup(.{
		.environment = .{
			.defaults = .{
				.color_format = .RGBA8,
				.depth_format = .DEPTH_STENCIL,
				.sample_count = 1,
			},
		},

		.logger = .{
			.func = sokol.log.func,
		},
	});

	defer {
		sokol.gfx.shutdown();
	}

	var pools = try Pools.init(coral.heap.allocator);

	defer {
		pools.deinit();
	}

	const Vertex = struct {
		xy: @Vector(2, f32),
		uv: @Vector(2, f32),
	};

	const quad_indices = [_]u16{0, 1, 2, 0, 2, 3};

	const quad_vertices = [_]Vertex{
		.{.xy = .{-0.5, -0.5}, .uv = .{0, 1}},
		.{.xy = .{0.5, -0.5}, .uv = .{1, 1}},
		.{.xy = .{0.5, 0.5}, .uv = .{1, 0}},
		.{.xy = .{-0.5, 0.5},  .uv = .{0, 0}},
	};

	quad_index_buffer = sokol.gfx.makeBuffer(.{
		.data = sokol.gfx.asRange(&quad_indices),
		.type = .INDEXBUFFER,
	});

	quad_vertex_buffer = sokol.gfx.makeBuffer(.{
		.data = sokol.gfx.asRange(&quad_vertices),
		.type = .VERTEXBUFFER,
	});

	view_buffer = sokol.gfx.makeBuffer(.{
		.data = sokol.gfx.asRange(&lina.orthographic_projection(-1.0, 1.0, .{
			.left = 0,
			.top = 0,
			.right = @floatFromInt(1280),
			.bottom = @floatFromInt(720),
		})),

		.type = .STORAGEBUFFER,
		.usage = .IMMUTABLE,
	});

	const shader_spirv = @embedFile("./shaders/2d_default.spv");
	var spirv_unit = try spirv.Unit.init(coral.heap.allocator);

	defer spirv_unit.deinit();

	try spirv_unit.compile(shader_spirv, .vertex);
	try spirv_unit.compile(shader_spirv, .fragment);

	Frame.texture_batching_pipeline = sokol.gfx.makePipeline(.{
		.label = "2D drawing pipeline",
		.layout = vertex_layout_state,
		.shader = sokol.gfx.makeShader(spirv_unit.shader_desc),
		.index_type = .UINT16,
	});

	var has_commands_head: ?*Commands = null;
	var has_commands_tail: ?*Commands = null;

	while (true) {
		switch (Work.pending.dequeue()) {
			.create_commands => |create_commands| {
				const commands = coral.heap.allocator.create(Commands) catch {
					_ = create_commands.created.resolve(error.OutOfMemory);

					continue;
				};

				commands.* = Commands.init(coral.heap.allocator);

				if (create_commands.created.resolve(commands)) {
					if (has_commands_tail) |tail_commands| {
						tail_commands.has_next = commands;
						commands.has_prev = tail_commands;
					} else {
						std.debug.assert(has_commands_head == null);

						has_commands_head = commands;
					}

					has_commands_tail = commands;
				} else {
					coral.heap.allocator.destroy(commands);
				}
			},

			.load_texture => |load| {
				const texture = try pools.create_texture(load.desc);

				if (!load.loaded.resolve(texture)) {
					std.debug.assert(pools.destroy_texture(texture));
				}
			},

			.render_frame => |render_frame| {
				var frame = Frame{
					.swapchain = .{
						.width = render_frame.width,
						.height = render_frame.height,
						.sample_count = 1,
						.color_format = .RGBA8,
						.depth_format = .DEPTH_STENCIL,
						.gl = .{.framebuffer = 0},
					}
				};

				sokol.gfx.beginPass(swapchain_pass: {
					var pass = sokol.gfx.Pass{
						.swapchain = frame.swapchain,

						.action = .{
							.stencil = .{.load_action = .CLEAR},
							.depth = .{.load_action = .CLEAR},
						},
					};

					pass.action.colors[0] = .{
						.clear_value = @bitCast(render_frame.clear_color),
						.load_action = .CLEAR,
					};

					break: swapchain_pass pass;
				});

				var has_commands = has_commands_head;

				while (has_commands) |commands| : (has_commands = commands.has_next) {
					for (commands.submitted_commands()) |command| {
						try command.process(&pools, &frame);
					}

					if (frame.current_target_texture != .default) {
						frame.set_target(&pools, .{
							.clear_color = null,
							.clear_depth = null,
							.clear_stencil = null,
						});
					}
				}

				frame.flush(&pools);
				sokol.gfx.endPass();
				sokol.gfx.commit();
				ext.SDL_GL_SwapWindow(window);
				render_frame.finished.set();
			},

			.rotate_commands => |rotate_commands| {
				var has_commands = has_commands_head;

				while (has_commands) |commands| : (has_commands = commands.has_next) {
					commands.rotate();
				}

				rotate_commands.finished.set();
			},

			.shutdown => {
				break;
			},

			.unload_texture => |unload| {
				if (!pools.destroy_texture(unload.handle)) {
					@panic("Attempt to unload a non-existent texture");
				}
			},
		}
	}
}

var work_thread: std.Thread = undefined;
