const coral = @import("coral");

const ext = @cImport({
	@cInclude("spirv-cross/spirv_cross_c.h");
});

const sokol = @import("sokol");

const std = @import("std");

pub const Error = std.mem.Allocator.Error || error {
	UnsupportedTarget,
	InvalidSPIRV,
	UnsupportedSPIRV,
	UnsupportedBackend,
};

pub const Unit = struct {
	arena: std.heap.ArenaAllocator,
	context: ext.spvc_context,
	shader_desc: sokol.gfx.ShaderDesc,
	attrs_used: u32 = 0,

	pub fn compile(self: *Unit, spirv_data: []const u8, stage: Stage) Error!void {
		if ((spirv_data.len % @alignOf(u32)) != 0) {
			return error.InvalidSPIRV;
		}

		const spirv_ops: []const u32 = @alignCast(std.mem.bytesAsSlice(u32, spirv_data));

		const execution_model, const stage_desc = switch (stage) {
			.vertex => .{ext.SpvExecutionModelVertex, &self.shader_desc.vs},
			.fragment => .{ext.SpvExecutionModelFragment, &self.shader_desc.fs},
		};

		const Backend = struct {
			target: ext.spvc_backend,
			option_values: []const struct {ext.spvc_compiler_option, c_uint},
		};

		const backend: Backend = try switch (sokol.gfx.queryBackend()) {
			.GLCORE => .{
				.target = ext.SPVC_BACKEND_GLSL,

				.option_values = &.{
					.{ext.SPVC_COMPILER_OPTION_GLSL_VERSION, 430},
					.{ext.SPVC_COMPILER_OPTION_GLSL_ES, @intFromBool(false)},
					.{ext.SPVC_COMPILER_OPTION_GLSL_VULKAN_SEMANTICS, @intFromBool(false)},
					.{ext.SPVC_COMPILER_OPTION_GLSL_EMIT_UNIFORM_BUFFER_AS_PLAIN_UNIFORMS, @intFromBool(true)},
					.{ext.SPVC_COMPILER_OPTION_FLIP_VERTEX_Y, @intFromBool(true)},
				},
			},

			else => error.UnsupportedBackend,
		};

		const compiler = parse_and_configure: {
			var parsed_ir: ext.spvc_parsed_ir = null;

			try to_error(ext.spvc_context_parse_spirv(self.context, spirv_ops.ptr, spirv_ops.len, &parsed_ir));

			var compiler: ext.spvc_compiler = null;

			try to_error(ext.spvc_context_create_compiler(self.context, backend.target, parsed_ir, ext.SPVC_CAPTURE_MODE_TAKE_OWNERSHIP, &compiler));
			try to_error(ext.spvc_compiler_build_combined_image_samplers(compiler));

			var combined_image_samplers: []const ext.spvc_combined_image_sampler = &.{};

			try to_error(ext.spvc_compiler_get_combined_image_samplers(compiler, @ptrCast(&combined_image_samplers.ptr), &combined_image_samplers.len));

			const arena_allocator = self.arena.allocator();
			var binding: u32 = 0;

			for (combined_image_samplers) |combined_image_sampler| {
				const name = try coral.utf8.alloc_formatted(arena_allocator, "{image_name}_{sampler_name}", .{
					.image_name = std.mem.span(@as([*:0]const u8, ext.spvc_compiler_get_name(compiler, combined_image_sampler.image_id))),
					.sampler_name = std.mem.span(@as([*:0]const u8, ext.spvc_compiler_get_name(compiler, combined_image_sampler.sampler_id))),
				});

				ext.spvc_compiler_set_name(compiler, combined_image_sampler.combined_id, name);
				ext.spvc_compiler_set_decoration(compiler, combined_image_sampler.combined_id, ext.SpvDecorationBinding, binding);

				binding += 1;
			}

			break: parse_and_configure compiler;
		};

		try to_error(ext.spvc_compiler_set_entry_point(compiler, stage_desc.entry, @intCast(execution_model)));

		const resources = create: {
			var resources: ext.spvc_resources = null;

			try to_error(ext.spvc_compiler_create_shader_resources(compiler, &resources));

			break: create resources;
		};

		try reflect_uniform_buffers(compiler, resources, stage_desc);
		try reflect_storage_buffers(compiler, resources, stage_desc);
		try reflect_image_samplers(compiler, resources, stage_desc);

		try to_error(ext.spvc_compiler_install_compiler_options(compiler, create: {
			var options: ext.spvc_compiler_options = null;

			try to_error(ext.spvc_compiler_create_compiler_options(compiler, &options));

			for (backend.option_values) |option_value| {
				const entry, const value = option_value;

				try to_error(ext.spvc_compiler_options_set_uint(options, entry, value));
			}

			break: create options;
		}));

		try to_error(ext.spvc_compiler_compile(compiler, @ptrCast(&stage_desc.source)));

		std.log.info("{s}", .{stage_desc.source});
	}

	pub fn deinit(self: *Unit) void {
		ext.spvc_context_destroy(self.context);
		self.arena.deinit();

		self.* = undefined;
	}

	pub fn init(allocator: std.mem.Allocator) std.mem.Allocator.Error!Unit {
		var context: ext.spvc_context = null;

		if (ext.spvc_context_create(&context) != ext.SPVC_SUCCESS) {
			return error.OutOfMemory;
		}

		errdefer ext.spvc_context_destroy(context);

		ext.spvc_context_set_error_callback(context, log_context_errors, null);

		return .{
			.arena = std.heap.ArenaAllocator.init(allocator),
			.context = context,

			.shader_desc = .{
				.vs = .{.entry = "main"},
				.fs = .{.entry = "main"},
			},
		};
	}
};

pub const Stage = enum {
	fragment,
	vertex,
};

fn log_context_errors(userdata: ?*anyopaque, error_message: [*c]const u8) callconv(.C) void {
	std.debug.assert(userdata == null);
	std.log.err("{s}", .{error_message});
}

fn reflect_image_samplers(compiler: ext.spvc_compiler, resources: ext.spvc_resources, stage_desc: *sokol.gfx.ShaderStageDesc) Error!void {
	var sampled_images: []const ext.spvc_reflected_resource = &.{};

	try to_error(ext.spvc_resources_get_resource_list_for_type(
		resources,
		ext.SPVC_RESOURCE_TYPE_SAMPLED_IMAGE,
		@ptrCast(&sampled_images.ptr),
		&sampled_images.len,
	));

	if (sampled_images.len > stage_desc.image_sampler_pairs.len) {
		return error.UnsupportedSPIRV;
	}

	for (0 .. sampled_images.len, sampled_images) |i, sampled_image| {
		const sampled_image_type = ext.spvc_compiler_get_type_handle(compiler, sampled_image.type_id);

		if (ext.spvc_type_get_basetype(sampled_image_type) != ext.SPVC_BASETYPE_SAMPLED_IMAGE) {
			return error.InvalidSPIRV;
		}

		stage_desc.images[i] = .{
			.multisampled = ext.spvc_type_get_image_multisampled(sampled_image_type) != 0,

			.image_type = try switch (ext.spvc_type_get_image_dimension(sampled_image_type)) {
				ext.SpvDim2D => sokol.gfx.ImageType._2D,
				else => error.InvalidSPIRV,
			},

			.sample_type = try switch (ext.spvc_type_get_basetype(ext.spvc_compiler_get_type_handle(compiler, ext.spvc_type_get_image_sampled_type(sampled_image_type)))) {
				ext.SPVC_BASETYPE_FP32 => sokol.gfx.ImageSampleType.FLOAT,
				else => error.InvalidSPIRV,
			},

			.used = true,
		};

		stage_desc.samplers[i] = .{
			.sampler_type = .DEFAULT,
			.used = true,
		};

		stage_desc.image_sampler_pairs[i] = .{
			.glsl_name = ext.spvc_compiler_get_name(compiler, sampled_image.id),
			.image_slot = @intCast(i),
			.sampler_slot = @intCast(i),
			.used = true,
		};
	}
}

fn reflect_storage_buffers(compiler: ext.spvc_compiler, resources: ext.spvc_resources, stage_desc: *sokol.gfx.ShaderStageDesc) Error!void {
	var storage_buffers: []const ext.spvc_reflected_resource = &.{};

	try to_error(ext.spvc_resources_get_resource_list_for_type(
		resources,
		ext.SPVC_RESOURCE_TYPE_STORAGE_BUFFER,
		@ptrCast(&storage_buffers.ptr),
		&storage_buffers.len,
	));

	for (storage_buffers) |storage_buffer| {
		const binding = ext.spvc_compiler_get_decoration(compiler, storage_buffer.id, ext.SpvDecorationBinding);

		if (binding >= stage_desc.storage_buffers.len) {
			return error.InvalidSPIRV;
		}

		var block_decorations: []const ext.SpvDecoration = &.{};

		try to_error(ext.spvc_compiler_get_buffer_block_decorations(
			compiler,
			storage_buffer.id,
			@ptrCast(&block_decorations.ptr),
			&block_decorations.len,
		));

		stage_desc.storage_buffers[binding] = .{
			.readonly = std.mem.indexOfScalar(ext.SpvDecoration, block_decorations, ext.SpvDecorationNonWritable) != null,
			.used = true,
		};
	}
}

fn reflect_uniform_buffers(compiler: ext.spvc_compiler, resources: ext.spvc_resources, stage_desc: *sokol.gfx.ShaderStageDesc) Error!void {
	var uniform_buffers: []const ext.spvc_reflected_resource = &.{};

	try to_error(ext.spvc_resources_get_resource_list_for_type(
		resources,
		ext.SPVC_RESOURCE_TYPE_UNIFORM_BUFFER,
		@ptrCast(&uniform_buffers.ptr),
		&uniform_buffers.len,
	));

	if (uniform_buffers.len != 0) {
		return error.InvalidSPIRV;
	}

	// TODO: Support for older APIs?
	_ = stage_desc;
	_ = compiler;
}

fn to_error(result: ext.spvc_result) Error!void {
	return switch (result) {
		ext.SPVC_SUCCESS => {},
		ext.SPVC_ERROR_INVALID_SPIRV => error.InvalidSPIRV,
		ext.SPVC_ERROR_UNSUPPORTED_SPIRV => error.UnsupportedSPIRV,
		ext.SPVC_ERROR_OUT_OF_MEMORY => error.OutOfMemory,
		ext.SPVC_ERROR_INVALID_ARGUMENT, ext.SPVC_ERROR_INT_MAX => unreachable,
		else => unreachable,
	};
}
